"""Docker Execution Module."""

import json
from typing import TypedDict, cast

import docker
from docker.errors import APIError

from opthub_runner.converter import float_to_json_float


class DockerConfig(TypedDict):
    """A type for docker execution configuration."""

    image: str
    environments: dict[str, str]
    command: list[str]
    timeout: float
    rm: bool


def execute_in_docker(
    config: DockerConfig,
    std_in: list[str],
) -> dict[str, object]:
    """Execute command in docker container.

    Args:
        config (DockerConfig): docker image name
        std_in (list[str]): standard input

    Returns:
        dict[str, Any]: parsed standard output
    """
    client = docker.from_env()

    try:
        client.images.pull(config["image"])  # pull image
    except APIError:
        client.images.get(config["image"])  # If image in local, get it
        
    container = client.containers.run(
        image=config["image"],
        command=config["command"],
        environment=config["environments"],
        stdin_open=True,
        detach=True,
    )

    socket = container.attach_socket(
        params={"stdin": 1, "stream": 1, "stdout": 1, "stderr": 1},
    )

    for line in std_in:
        socket._sock.sendall(line.encode("utf-8"))  # noqa: SLF001

    container.wait(timeout=config["timeout"])

    stdout = container.logs(stdout=True, stderr=False).decode("utf-8")

    if config["rm"]:
        container.remove()

    out: dict[str, object] | None = parse_stdout(stdout)

    if out is None:
        msg = "Failed to parse stdout."
        raise RuntimeError(msg)

    return cast(dict[str, object], float_to_json_float(out))


def parse_stdout(stdout: str) -> dict[str, object] | None:
    """Parse stdout.

    Args:
        stdout (str): stdout

    Returns:
        dict[str, Any] | None: parsed stdout
    """
    lines = stdout.split("\n")
    lines.reverse()
    for line in lines:
        if line:
            line_dict: dict[str, object] = json.loads(line)
            return line_dict
    return None



"""This module implements the evaluator."""

import json
from typing import TypedDict

from opthub_runner.docker_executor import execute_in_docker


class EvaluationResult(TypedDict):
    """A type for evaluation result."""

    objective: object
    feasible: bool | None
    constraint: object | None
    info: object | None


class Evaluator:
    """The Evaluator class."""

    def __init__(
        self,
        docker_image: str,
        environment: dict[str, object],
        *,
        rm: bool = True,
        timeout: float = 43200,
    ) -> None:
        """Initialize the Evaluator class.

        Args:
            docker_image (str): The docker image URL.
            environment (dict[str, object]): The environments.
            rm (bool, optional):
                Remove the container after execution. Defaults to True.
            timeout (float, optional):
                The timeout for the execution. Defaults to 43200.
        """
        self.docker_image = docker_image
        self.timeout = timeout
        self.rm = rm

        # Convert all environment variables to string
        self.environment: dict[str, str] = {var: str(environment[var]) for var in environment}

    def run(self, variable: object) -> EvaluationResult:
        """Run the evaluator.

        Args:
            variable (object): The variable to evaluate.

        Returns:
            EvaluationResult: The evaluation result.
        """
        docker_output = execute_in_docker(
            {
                "image": self.docker_image,
                "environments": self.environment,
                "command": [],
                "timeout": self.timeout,
                "rm": self.rm,
            },
            [json.dumps(variable) + "\n"],
        )
        if "error" in docker_output:
            error = docker_output["error"]
            msg = f"Error occurred while evaluating solution:\n{error}"
            raise RuntimeError(msg)

        feasible = docker_output.get("feasible", None)

        if feasible is not None and not isinstance(feasible, bool):
            msg = f"Feasible value is not boolean: {feasible}"
            raise TypeError(msg)

        evaluation_result: EvaluationResult = {
            "objective": docker_output["objective"],
            "feasible": feasible,
            "constraint": docker_output.get("constraint", None),
            "info": docker_output.get("info", None),
        }
        return evaluation_result



"""This module contains the functions to convert."""

import logging
import math
import sys

LOGGER = logging.getLogger(__name__)


def float_to_json_float(value: object) -> object:  # noqa: PLR0911
    """Convert float values to JSON float values.

    Args:
        value (object): The object consists of float values.

    Returns:
        object: The object consists of json float values.
    """
    if isinstance(value, list):
        return [float_to_json_float(v) for v in value]
    if isinstance(value, dict):
        return {k: float_to_json_float(v) for k, v in value.items()}
    if not isinstance(value, float):
        return value
    if value == math.inf:
        LOGGER.warning("math.inf is converted to sys.float_info.max")
        return sys.float_info.max
    if value == -math.inf:
        LOGGER.warning("-math.inf is converted to -sys.float_info.max")
        return -sys.float_info.max
    if math.isnan(value):
        LOGGER.warning("math.nan is converted to None")
        return None
    return value
